/* Audio Library for Teensy 3.X
 * Copyright (c) 2014, Paul Stoffregen, paul@pjrc.com
 *
 * Development of this audio library was funded by PJRC.COM, LLC by sales of
 * Teensy and Audio Adaptor boards.  Please support PJRC's efforts to develop
 * open source software by purchasing Teensy or other PJRC products.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice, development funding notice, and this permission
 * notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifndef synth_sine_h_
#define synth_sine_h_

#include "Arduino.h"
#include "AudioStream.h"
#include "arm_math.h"

// TODO: investigate making a high resolution sine wave
// using Taylor series expansion.
// http://www.musicdsp.org/showone.php?id=13


class AudioSynthWaveformSine : public AudioStream
{
public:
	AudioSynthWaveformSine() : AudioStream(0, NULL), magnitude(16384) {}
	void frequency(float freq) {
		if (freq < 0.0) freq = 0.0;
		else if (freq > AUDIO_SAMPLE_RATE_EXACT/2) freq = AUDIO_SAMPLE_RATE_EXACT/2;
		phase_increment = freq * (4294967296.0 / AUDIO_SAMPLE_RATE_EXACT);
	}
	void phase(float angle) {
		if (angle < 0.0) angle = 0.0;
		else if (angle > 360.0) {
			angle = angle - 360.0;
			if (angle >= 360.0) return;
		}
		phase_accumulator = angle * (4294967296.0 / 360.0);
	}
	void amplitude(float n) {
		if (n < 0) n = 0;
		else if (n > 1.0) n = 1.0;
		magnitude = n * 65536.0;
	}
	virtual void update(void);
private:
	uint32_t phase_accumulator;
	uint32_t phase_increment;
	int32_t magnitude;
};


class AudioSynthWaveformSineHires : public AudioStream
{
public:
	AudioSynthWaveformSineHires() : AudioStream(0, NULL), magnitude(16384) {}
	void frequency(float freq) {
		if (freq < 0.0) freq = 0.0;
		else if (freq > AUDIO_SAMPLE_RATE_EXACT/2) freq = AUDIO_SAMPLE_RATE_EXACT/2;
		phase_increment = freq * (4294967296.0 / AUDIO_SAMPLE_RATE_EXACT);
	}
	void phase(float angle) {
		if (angle < 0.0) angle = 0.0;
		else if (angle > 360.0) {
			angle = angle - 360.0;
			if (angle >= 360.0) return;
		}
		phase_accumulator = angle * (4294967296.0 / 360.0);
	}
	void amplitude(float n) {
		if (n < 0) n = 0;
		else if (n > 1.0) n = 1.0;
		magnitude = n * 65536.0;
	}
	virtual void update(void);
private:
	uint32_t phase_accumulator;
	uint32_t phase_increment;
	int32_t magnitude;
};


class AudioSynthWaveformSineModulated : public AudioStream
{
public:
	AudioSynthWaveformSineModulated() : AudioStream(1, inputQueueArray), magnitude(16384) {}
	// maximum unmodulated carrier frequency is 11025 Hz
	// input = +1.0 doubles carrier
	// input = -1.0 DC output
	void frequency(float freq) {
		if (freq < 0.0) freq = 0.0;
		else if (freq > AUDIO_SAMPLE_RATE_EXACT/4) freq = AUDIO_SAMPLE_RATE_EXACT/4;
		phase_increment = freq * (4294967296.0 / AUDIO_SAMPLE_RATE_EXACT);
	}
	void phase(float angle) {
		if (angle < 0.0) angle = 0.0;
		else if (angle > 360.0) {
			angle = angle - 360.0;
			if (angle >= 360.0) return;
		}
		phase_accumulator = angle * (4294967296.0 / 360.0);
	}
	void amplitude(float n) {
		if (n < 0) n = 0;
		else if (n > 1.0) n = 1.0;
		magnitude = n * 65536.0;
	}
	virtual void update(void);
private:
	uint32_t phase_accumulator;
	uint32_t phase_increment;
	audio_block_t *inputQueueArray[1];
	int32_t magnitude;
};
class Envelope 
{
public:
	Envelope() :  noteon(0),segments(0){
	}
	// immediately jump to the new DC level
	void noteOn(int fx,float dur){
		#if 0
		Serial.print(noteon);
		Serial.print(" on ");
		Serial.print(time-count);
		Serial.print(" ");
		Serial.println(prev);
		#endif
		actseg=0;
		count=0;
		noteon=1;
		prev =levels[actseg];
//		Serial.println(fx);
	}

	void noteOff(float n, float milliseconds){
		actseg=segments-2;
		count=0;
		noteon=1;
	}
	void multiseg(float *plevels,float *pslope,int32_t  psegments,float *ptimes,float *pamps)
	{
		levels=plevels;
		segments=psegments;
		slope=pslope;
		amps=pamps;
		times=ptimes;
	}
	void multiseg(int s,float pamps)
	{
		amps[s]=pamps;
		Serial.println(amps[s]);
	}
	float update(bool isSolo,int loopend,int loopmark);
	int noteon;
	int32_t  segments;
	bool show=false;
	int32_t  count; // sample count;
	float lastm=0;
	float *amps;
	bool solo=false;
	bool mute=false;
    float af=1.0;
	float tf=1.0;
	bool debug=false;
private:
	float *slope;
	float *levels;
	float *times;
	int32_t  actseg;
	float prev;
};
class AudioSynthWaveformHarmModulated : public AudioStream
{
public:
	AudioSynthWaveformHarmModulated() : AudioStream(1, inputQueueArray){
		for(int i=0;i<32;i++)
		{
			magnitude[i]=16384;
		}
		nfreq=1;
		ismulti=false;
	}
	void partials(int nf) {
		if(nf>32)
			nf=32;
		nfreq=nf;
//				Serial.print("partials ");
//				Serial.println(nfreq);
	}
	void segments(int ns) {
		if(ns>128)
			ns=128;
		nseq=ns;
		for(int i=0;i<nfreq;i++)
		{
			envelope[i].segments=nseq;
		}
//		Serial.println(nseq);
	}
	// maximum unmodulated carrier frequency is 11025 Hz
	// input = +1.0 doubles carrier
	// input = -1.0 DC output
	void frequency(float *freqin) {
		for(int i=0;i<nfreq;i++)
		{
			float freq=freqin[i];
			if (freq < 0.0) freq = 0.0;
			else if (freq > AUDIO_SAMPLE_RATE_EXACT/4) freq = AUDIO_SAMPLE_RATE_EXACT/4;
			phase_increment[i] = freq * (4294967296.0 / AUDIO_SAMPLE_RATE_EXACT);
			phase_accumulator[i] = 0;
		}
	}
	void frequency(float freqin) {
//			Serial.print("nfreq ");
//			Serial.print(nfreq);
//			Serial.print(" ");
		for(int i=0;i<nfreq;i++)
		{
			float freq=freqin*(i+1);
			if(freqlist!=0)
				freq=freqin*freqlist[i]+offlist[i];
			if (freq < 0.0) freq = 0.0;
			else if (freq > AUDIO_SAMPLE_RATE_EXACT/4) freq = AUDIO_SAMPLE_RATE_EXACT/4;
			phase_increment[i] = freq * (4294967296.0 / AUDIO_SAMPLE_RATE_EXACT);
			phase_accumulator[i] = phaselist[i]* (4294967296.0 / 360.0);
#if 0
			Serial.print(freqlist[i]);
			Serial.print(" ");
			Serial.print(offlist[i]);
			Serial.print(" ");
			Serial.print(phaselist[i]);
			Serial.print(" ");
#endif
		}
//		Serial.println();
	}
	void phase(float *anglein) {
		for(int i=0;i<nfreq;i++)
		{
			float angle=anglein[i];
			if (angle < 0.0) angle = 0.0;
			else if (angle > 360.0) {
				angle = angle - 360.0;
				if (angle >= 360.0) continue;
			}
			phase_accumulator[i] = angle * (4294967296.0 / 360.0);
		}
	}
	void amplitude(float *nin) {
		for(int i=0;i<nfreq;i++)
		{
			float n=nin[i];
			if (n < 0) n = 0;
			else if (n > 1.0) n = 1.0;
			magnitude[i] = n * 65536.0;
			phase_accumulator[i] = phaselist[i]* (4294967296.0 / 360.0);
		}
	}
#if 0
	void setfmax(float n,int omni,int part) {
		int fs=part,fe=part+1;
		if(omni)
		{
			fs=0;
			fe=nfreq;
		}
//		for(int i=fs;i<fe;i++)
//			amps[i]=n;
//		Serial.println(fmax);
	}
#endif
	void amplitude(float n) {
		for(int i=0;i<nfreq;i++)
		{
			if (n < 0) n = 0;
			else if (n > 1.0) n = 1.0;
			magnitude[i] = n * 65536.0;
			phase_accumulator[i] = phaselist[i]* (4294967296.0 / 360.0);
		}
	}
	void setFreq(float *pfreqlist, float *pofflist, float *pphaselist)
	{
			freqlist=pfreqlist;
			offlist=pofflist;
			phaselist=pphaselist;
//			Serial.print("set f");
//			Serial.println(freqlist[1]);
	}
	void noteOn(int fx,float dur){
		noteon=true;

		for(int i=0;i<nfreq;i++)
		{
			envelope[i].noteOn(fx,dur);
		}
	}
	void noteOff(float n, float milliseconds){
		noteon=false;
		for(int i=0;i<nfreq;i++)
		{
			envelope[i].noteOff(n, milliseconds);
		}
	}
	void multiseg(float *plevels,float *pslope,int32_t  psegments,float *pphase,int p,float *ptimes,float *pamps)
	{
		phase_accumulator[p] = pphase[p] * (4294967296.0 / 360.0);
		envelope[p].multiseg(plevels,pslope,psegments,ptimes,pamps);
		ismulti=true;
		m_time=ptimes[0];
	}
	void multiseg(int p,int s,float val)
	{
		envelope[p].multiseg(s,val);
		Serial.print("multi ");
		Serial.print(p);
		Serial.print(" , ");
		Serial.print(s);
		Serial.print(" = ");
		Serial.println(envelope[p].amps[s]);
	}
	void setshow(int part)
	{
		if(part>-1)
			envelope[part].show=true;
		else
			envelope[-part].show=false;
	}
	void setsolo(int part,bool set)
	{
			envelope[part].solo=set;
			if(set)
				solos++;
			else
				solos--;
	}
	void setmute(int part,bool mute)
	{
			envelope[part].mute=mute;
	}

	virtual void update(void);
	int nfreq;
	int nseq;
	bool noteon=false;
	float m_time;
	Envelope envelope[32];
	int solos=0;
	int loopmark=-1;
	int loopend=-1;
private:
	float *freqlist;
	float *offlist;
	float *phaselist;
	uint32_t phase_accumulator[32];
	uint32_t phase_increment[32];
	audio_block_t *inputQueueArray[1];
	int32_t magnitude[32];
    bool ismulti;
};

#endif
