/* 
 * AdioSynthWaveformSine_F32
 * 
 * Created: Chip Audette (OpenAudio) Feb 2017
 * Modeled on: AudioSynthWaveformSine from Teensy Audio Library
 * 
 * Purpose: Create sine wave of given amplitude and frequency
 *
 * License: MIT License. Use at your own risk.        
 *
 */

#ifndef synth_sine_f32_h_
#define synth_sine_f32_h_

#include "Arduino.h"
#include "AudioStream_F32.h"
#include "arm_math.h"


class AudioSynthWaveformSine_F32 : public AudioStream_F32
{
//GUI: inputs:0, outputs:1 //this line used for automatic generation of GUI node
//GUI: shortName:sine  //this line used for automatic generation of GUI node
public:
	AudioSynthWaveformSine_F32() : AudioStream_F32(0, NULL), magnitude(16384) { } //uses default AUDIO_SAMPLE_RATE from AudioStream.h
	AudioSynthWaveformSine_F32(const AudioSettings_F32 &settings) : AudioStream_F32(0, NULL), magnitude(16384) {
		setSampleRate_Hz(settings.sample_rate_Hz);
	}
	void frequency(float freq) {
		if (freq < 0.0) freq = 0.0;
		else if (freq > sample_rate_Hz/2.f) freq = sample_rate_Hz/2.f;
		phase_increment = freq * (4294967296.0 / sample_rate_Hz);
	}
	void phase(float angle) {
		if (angle < 0.0f) angle = 0.0f;
		else if (angle > 360.0f) {
			angle = angle - 360.0f;
			if (angle >= 360.0f) return;
		}
		phase_accumulator = angle * (4294967296.0f / 360.0f);
	}
	void amplitude(float n) {
		if (n < 0) n = 0;
		else if (n > 1.0f) n = 1.0f;
		magnitude = n * 65536.0f;
	}
	void setSampleRate_Hz(const float &fs_Hz) {
		phase_increment *= sample_rate_Hz / fs_Hz; //change the phase increment for the new frequency
		sample_rate_Hz = fs_Hz;
	}
	void begin(void) { enabled = true; }
	void end(void) { enabled = false; }
	virtual void update(void);
	
private:
	uint32_t phase_accumulator = 0;
	uint32_t phase_increment = 0;
	int32_t magnitude = 0;
	float sample_rate_Hz = AUDIO_SAMPLE_RATE;
	volatile uint8_t enabled = 1;
};
class Envelope_F32 
{
public:
	Envelope_F32 () :  noteon(false),state(0), magnitude(0),segments(0){
	}

	void noteOn(int fx,float dur){
		actseg=0;
		count=0;
		noteon=true;
		showfx=fx;
		duration=dur;
		rescount=0;
		prev =levels[actseg];
		float tau = 0.9;
		a = exp(-1/(duration*times[actseg]*tau));
	}

	void noteOff(float n, float milliseconds){
		actseg=0;
		count=0;
		noteon=false;
	}
	void multiseg(float *plevels,float *ptimes,float *pslope,int32_t  psegments)
	{
		times=ptimes;
		levels=plevels;
		segments=psegments;
		slope=pslope;
	}
	float update(int fx);
	bool noteon;

private:
	uint8_t  state;     // 0=steady output, 1=transitioning
	int32_t  magnitude; // current output
	int32_t  target;    // designed output (while transitiong)
	int32_t  increment; // adjustment per sample (while transitiong)
	int32_t  count; // sample count;
	float *slope;
	float *times;
	float *levels;
	float maxlevel;
	int32_t  segments;
	int32_t  actseg;
	int showfx;
	float duration;
	int rescount;
	float *eresult;
	float prev;
	float a;
};
//GUI: inputs:0, outputs:1 //this line used for automatic generation of GUI node
//GUI: shortName:sineharmf32  //this line used for automatic generation of GUI node

class AudioSynthWaveformHarm_F32 : public AudioStream_F32
{
public:
	AudioSynthWaveformHarm_F32() : AudioStream_F32(0, NULL){
		for(int i=0;i<32;i++)
		{
			magnitude[i]=16384;
		}
		nfreq=1;
		ismulti=false;
	}
	void partials(int nf) {
		if(nf>32)
			nf=32;
		nfreq=nf;
	}
	void frequency(float *freqin) {
		for(int i=0;i<nfreq;i++)
		{
			float freq=freqin[i];
			if (freq < 0.0) freq = 0.0;
			else if (freq > AUDIO_SAMPLE_RATE_EXACT/4) freq = AUDIO_SAMPLE_RATE_EXACT/4;
			phase_increment[i] = freq * (4294967296.0 / AUDIO_SAMPLE_RATE_EXACT);
			phase_accumulator[i] = 0;
		}
	}
	void frequency(float freqin) {
		for(int i=0;i<nfreq;i++)
		{
			float freq=freqin*(i+1);
			if(freqlist!=0)
				freq=freqin*freqlist[i]+offlist[i];
			if (freq < 0.0) freq = 0.0;
			else if (freq > AUDIO_SAMPLE_RATE_EXACT/2.f) freq = AUDIO_SAMPLE_RATE_EXACT/4;
			phase_increment[i] = freq * (4294967296.0 / AUDIO_SAMPLE_RATE_EXACT);
			phase_accumulator[i] = 0;
		}
	}
	void phase(float *anglein) {
		for(int i=0;i<nfreq;i++)
		{
			float angle=anglein[i];
			if (angle < 0.0) angle = 0.0;
			else if (angle > 360.0) {
				angle = angle - 360.0;
				if (angle >= 360.0) continue;
			}
			phase_accumulator[i] = angle * (4294967296.0 / 360.0);
		}
	}
	void amplitude(float *nin) {
		for(int i=0;i<nfreq;i++)
		{
			float n=nin[i];
			if (n < 0) n = 0;
			else if (n > 1.0) n = 1.0;
			magnitude[i] = n * 65536.0;
			phase_accumulator[i] = 0;
		}
	}
	void amplitude(float n) {
		for(int i=0;i<nfreq;i++)
		{
			if (n < 0) n = 0;
			else if (n > 1.0) n = 1.0;
			magnitude[i] = n * 65536.0;
			phase_accumulator[i] = 0;
		}
	}
	void setFreq(float *pfreqlist, float *pofflist)
	{
		for(int i=0;i <nfreq;i++)
		{
			freqlist[i]=pfreqlist[i];
			offlist[i]=pofflist[i];
		}
	}
	void noteOn(int fx,float dur){
		for(int i=0;i<nfreq;i++)
		{
			envelope[i].noteOn(fx,dur);
		}
	}
	void noteOff(float n, float milliseconds){
		for(int i=0;i<nfreq;i++)
		{
			envelope[i].noteOff(n, milliseconds);
		}
	}
	bool ismulri(void)
	{
		return(ismulti);
	}
	void multiseg(float *plevels,float *ptimes,float *pslope,int32_t  psegments,float phase,int p)
	{
		phase_accumulator[p] = phase * (4294967296.0 / 360.0);
		envelope[p].multiseg(plevels,ptimes,pslope,psegments);
		ismulti=true;
	}
	void begin(void) { enabled = true; }
	void end(void) { enabled = false; }

	virtual void update(void);
	int nfreq;
private:
	float freqlist[32];
	float offlist[32];
	uint32_t phase_accumulator[32];
	uint32_t phase_increment[32];
	int32_t magnitude[32];
	Envelope_F32  envelope[32];
    bool ismulti;
	volatile uint8_t enabled = 1;
};



#endif
